from subprocess import Popen

from .config import Config
from .daemon import Daemon, KeyEvent


def start():
    config = Config(use_default=True)

    def handle(event: KeyEvent):
        command = config.get_command(event)
        if command is not None:
            Popen(command, shell=True)

    daemon = Daemon(config.device_id)
    daemon.handlers.add(handle)
    daemon.start()



import os
from pathlib import Path
from shutil import copyfile
from typing import Dict, TypeVar

import yaml
from xdg import BaseDirectory

from pymacropad.daemon import KeyEvent, KeyState


class ConfigException(Exception):
    pass


class KeyConfig:
    def __init__(self, key: str, data: Dict[str, str] | str):
        self.key = key
        if type(data) == str:
            self.down = data
            self.held = None
            self.up = None
        else:
            self.down = data.get('down', None)
            self.held = data.get('held', None)
            self.up = data.get('up', None)

    def get_command(self, state: KeyState):
        if state == KeyState.DOWN:
            return self.down
        if state == KeyState.HELD:
            return self.held
        if state == KeyState.UP:
            return self.up
        return None

KC = TypeVar('KC', bound=KeyConfig)


class Config:
    _key_configs: Dict[str, KC] = {}
    _device_id: str
    _last_modified: float = None

    def __init__(self, config_file_name='pymacropad.yaml', use_default=False):
        self.config_file_name = config_file_name
        self.config_location = self._find_config() if not use_default else Config._default_config_path()
        print(f"Using config file at {self.config_location}")

    def get_command(self, event: KeyEvent):
        key_config = self.key_configs.get(event.key)
        if key_config is None:
            return
        return key_config.get_command(event.state)

    def _load(self):
        if os.path.getmtime(self.config_location) != self._last_modified:
            with open(self.config_location, 'r') as f:
                data = yaml.safe_load(f)
            self._last_modified = os.path.getmtime(self.config_location)

            self._key_configs = {k: self._build_key_config(k, v) for k, v in data.get('keys', {}).items()}
            self._device_id = data.get('device_id')

    @classmethod
    def _build_key_config(cls, key: str, data: dict) -> KC:
        return KeyConfig(key, data)

    @property
    def key_configs(self):
        self._load()
        return self._key_configs

    @property
    def device_id(self):
        self._load()
        return self._device_id

    def _find_config(self):
        config_base = Path(BaseDirectory.xdg_config_home)
        config_path = config_base.joinpath(self.config_file_name)
        if config_path.exists():
            return config_path
        return Config._create_config(config_path)

    @staticmethod
    def _create_config(config_location: Path):
        print(f"Creating default config file at {config_location}")
        copyfile(Config._default_config_path(), config_location)
        return config_location

    @staticmethod
    def _default_config_path():
        return Path(os.path.realpath(__file__)).parent.joinpath('default.yaml')



from enum import Enum
from typing import Callable, List, Set

import libevdev
from os import path

from libevdev import InputEvent


class KeyState(Enum):
    UP = 0
    DOWN = 1
    HELD = 2


class KeyEvent:
    @classmethod
    def from_event(cls, event: InputEvent, device_id: str):
        return cls(event.code.name, KeyState(event.value), device_id)

    def __init__(self, key: str, state: KeyState, device_id: str):
        self.key = key
        self.state = state
        self.device_id = device_id


class Daemon:
    def __init__(self, device_id: str):
        self.device_id = device_id
        self.handlers: Set[Callable[[KeyEvent], None]] = set()

    @property
    def device_path(self):
        return f'/dev/input/by-id/{self.device_id}'

    def start(self):
        while True:
            if path.exists(self.device_path):
                self._start_loop()


    def _start_loop(self):
        with open(self.device_path, 'rb') as fd:
            device = libevdev.Device(fd)
            device.grab()
            print("Device connected")

            while True:
                for event in device.events():
                    if not event.matches(libevdev.EV_KEY):
                        continue
                    self._handle(event)

    def _handle(self, event: InputEvent):
        key_event = KeyEvent.from_event(event, self.device_id)
        print(f"Handling key: {key_event.key}, state: {key_event.state}")

        for handler in self.handlers:
            handler(key_event)



if __name__ == '__main__':
    d = Daemon('usb-CapsUnlocked_CU7-event-kbd')
    d.start()



# class Daemon:
#     def __init__(self, config: Config):
#         self.config = config
#
#     def start(self):
#         while True:
#             if path.exists(self.config.device):
#                 self.handle_loop()
#             else:
#                 sleep(3)
#
#     def handle_loop(self):
#         try:
#             dev = InputDevice(self.config.device)
#             print("device connected")
#
#             with dev.grab_context():
#                 while True:
#                     select([dev], [], [])
#                     for event in dev.read():
#                         if event.type == ecodes.EV_KEY:
#                             self.handle(event.code, event.value)
#         except OSError as e:
#             print("device disconnected", e)
#
#     def handle(self, code, state):
#         key, mod = constants.SHIFT_MAP[code]
#         state = constants.STATE_MAP[state]
#         print(f"handling key: {key}, mod: {mod}, state: {state}")
#         conf = self.config.key_configs.get(key)
#         if conf is None:
#             return
#         command = conf.get_command(state, mod)
#         if command is None:
#             return
#         print(f"\texecuting: {command}")
#         Popen(command, shell=True)
#
#
# def start():
#     config = Config()
#     daemon = Daemon(config)
#     daemon.start()



from setuptools import setup, find_packages

setup(
    name='pymacropad',
    version_config={
        "template": "{tag}",
        "dev_template": "{tag}.dev{ccount}+git.{sha}",
        "dirty_template": "{tag}.dev{ccount}+git.{sha}.dirty",
        "starting_version": "0.0.1",
        "version_file": "",
        "count_commits_from_version_file": False
    },
    author="James Waters",
    author_email="james@jcwaters.co.uk",
    description="Utility program for binding actions to keys on a given input",
    url="https://github.com/j-waters/pymacropad",
    install_requires=['PyYAML', 'Click', 'libevdev', 'pyxdg'],
    setup_requires=['setuptools-git-versioning'],
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'pymacropad=pymacropad.__main__:main'
        ]
    },
    python_requires='>=3.10',
)
